/*
    (C) 1995-96 AROS - The Amiga Research OS
    $Id: readargs.c,v 1.3 2007/04/15 22:21:05 piru Exp $

    Desc:
    Lang: english
*/
#include <exec/memory.h>
#include <proto/exec.h>
#include <dos/rdargs.h>
#include <dos/dosextens.h>
#include "dos_intern.h"

#ifdef TEST
#undef __NOLIBBASE__
#undef SysBase
#undef UtilityBase
#define ReadArgs yyy
#    include <stdio.h>
#    include <proto/dos.h>
#    undef ReadArgs
#    undef AROS_LH3
#    define AROS_LH3(t,fn,a1,a2,a3,bt,bn,o,lib)     t fn (a1,a2,a3)
#    undef AROS_LHA
#    define AROS_LHA(t,n,r)                   t n
#    undef AROS_LIBFUNC_INIT
#    define AROS_LIBFUNC_INIT
#    undef AROS_LIBBASE_EXT_DECL
#    define AROS_LIBBASE_EXT_DECL(bt,bn)
#    undef AROS_LIBFUNC_EXIT
#    define AROS_LIBFUNC_EXIT
#endif

/*****************************************************************************

    NAME */
#include <proto/dos.h>

	AROS_LH3(struct RDArgs *, ReadArgs,

/*  SYNOPSIS */
	AROS_LHA(STRPTR,          template, D1),
	AROS_LHA(IPTR *,          array,    D2),
	AROS_LHA(struct RDArgs *, rdargs,   D3),

/*  LOCATION */
	struct DosLibrary *, DOSBase, 133, Dos)

/*  FUNCTION
	Parses the commandline, a given string or Input() and fills
	an argument array according to the options template given.
	The array must be initialized to the wanted defaults before
	each call to ReadArgs(). If the rdargs argument is NULL
	ReadArgs() tries to parse the commandline and continues
	on the input channel if it just consists of a single '?',
	prompting the user for input.

    INPUTS
	template - Template string. The template string is given as
		   a number of options separated by ',' and modified
		   by '/' modifiers, e.g. 'NAME,WIDTH/N,HEIGHT/N'
		   means get a name string and two numbers (width and
		   height). The possible modifiers are:
		   /S Option is a switch. It may be either set or
		      left out.
		   /T Option is a boolean value. Requires an argument
		      which may be "ON", "YES" (setting the respective
		      argument to 1), "OFF" or "NO" (setting the
		      respective argument to 0).
		   /N Option is a number. Strings are not allowed.
		      If the option is optional, a pointer to the
		      actual number is returned. This is how you know
		      if it was really given.
		   /A Argument is required. If it is left out ReadArgs()
		      fails.
		   /K The keyword must be given when filling the option.
		      Normally it's skipped.
		   /M Multiple strings. The result is returned as a string
		      pointer array terminated with NULL. /M eats all strings
		      that don't fit into any other option. If there are
		      unfilled /A arguments after parsing they steal strings
		      from /M. This makes it possible to e.g. write a COPY
		      template like 'FROM/A/M,TO/A'. There may be only one
		      /M option in a template.
		   /F Eats the rest of the line even if there are option
		      keywords in it.
	array	 - Array to be filled with the result values. The array must
		   be intialized to the default values before calling
		   ReadArgs().
	rdargs	 - An optional RDArgs structure determinating the type of
		   input to process.

    RESULT
	A handle for the memory allocated by ReadArgs(). Must be freed
	with FreeArgs() later.

    NOTES

    EXAMPLE
	See below.

    SEE ALSO
	FreeArgs(), Input()

    INTERNALS

    HISTORY
	29-10-95    digulla automatically created from
			    dos_lib.fd and clib/dos_protos.h

*****************************************************************************/
{
    AROS_LIBFUNC_INIT
    AROS_LIBBASE_EXT_DECL(struct DosLibrary *,DOSBase)

    /* Allocated resources */
    struct RDArgs *rda = NULL;
    struct DAList *dalist = NULL;
    UBYTE         *flags = NULL;
    STRPTR         strbuf = NULL, iline = NULL;
    STRPTR        *multvec = NULL, *argbuf = NULL;
    ULONG          multnum = 0, multmax = 0;

    /* Some variables */
    LONG           c;
    STRPTR         s1, s2, *newmult;
    ULONG          arg, numargs, nextarg;
    LONG           it, item, chars, value;
    struct CSource lcs, *cs = NULL;

    /* Get pointer to process structure. */
    struct Process *me = (struct Process *)FindTask(NULL);

    /* Error recovery. C has no exceptions. This is a simple replacement. */
    LONG error;
#undef ERROR
#define ERROR(a) { error=a; goto end; }

    /* Template options */
#define REQUIRED 0x80 /* /A */
#define KEYWORD  0x40 /* /K */
#define MULTIPLE 0x20 /* /M */
#define TYPEMASK 0x07
#define NORMAL	 0x00 /* No option */
#define SWITCH	 0x01 /* /S, implies /K */
#define TOGGLE	 0x02 /* /T, implies /K */
#define NUMERIC  0x03 /* /N */
#define REST	 0x04 /* /F */

#define GETCHAR(cs)				\
(cs) ?						\
    ((cs)->CS_CurChr < (cs)->CS_Length) ?	\
	(cs)->CS_Buffer[(cs)->CS_CurChr++]	\
    :						\
	EOF					\
:						\
    EOF

    /* Flags for each possible character. */
    static const UBYTE argflags[] =
    {
	REQUIRED, 0, 0, 0, 0, REST, 0, 0, 0, 0, KEYWORD, 0, MULTIPLE,
	NUMERIC, 0, 0, 0, 0, SWITCH|KEYWORD, TOGGLE|KEYWORD, 0, 0, 0, 0, 0, 0
    };

    DEBUG_RDARGS(dprintf("ReadArgs: template = \"%s\", array = $%lx, rda = $%lx, source = %s\n",
			 template, (ULONG) array, (ULONG) rdargs, (rdargs && rdargs->RDA_Source.CS_Buffer) ? "string" : "stdin"));

    FAILFORTASK(me, return NULL;)

    /* Allocate readargs structure (and private internal one) */
    rda = rdargs ? rdargs : (struct RDArgs *) dosAllocVec(sizeof(struct RDArgs), MEMF_ANY);
    dalist = (struct DAList *) dosAllocVec(sizeof(struct DAList), MEMF_ANY);

    if (!rda || !dalist)
	ERROR(ERROR_NO_FREE_STORE);

    dalist->FreeRDA = rdargs == NULL;

    /* Init character source. */
    if (rdargs != NULL && rdargs->RDA_Source.CS_Buffer)
    {
	cs = &rdargs->RDA_Source;
    }
    else
    {
	iline = internal_ReadLine(&lcs, DOSBase);
	if (!iline)
	    ERROR(me->pr_Result2);
	lcs.CS_CurChr = 0;
	cs = &lcs;
    }

    DEBUG_RDARGS(dprintf("ReadArgs: line = \"%s\"\n", &cs->CS_Buffer[cs->CS_CurChr]));

    /* Check for optional reprompting */
    if (!rdargs || !(rdargs->RDA_Flags & RDAF_NOPROMPT))
    {
	for (;;)
	{
	    BPTR  output;

	    /* Check commandline for a single '?' */
	    s1 = cs->CS_Buffer;

	    /* Skip leading whitespace */
	    while (*s1 == ' ' || *s1 == '\t')
		s1++;

	    /* Check for '?' */
	    if (*s1++ != '?')
		break;

	    /* Skip (one, yes only one!) whitespace */
	    if (*s1 == ' ' || *s1 == '\t')
		s1++;

	    /* Check for EOL */
	    if (*s1 != '\n' && *s1)
		break;

	    /* Only a single '?' on the commandline. */
	    output = Output();

	    /* Prompt for more input */
	    if (rdargs && rdargs->RDA_Flags & (1 << 30) && rdargs->RDA_ExtHelp)
	    {
	        s1 = rdargs->RDA_ExtHelp;
	    }
	    else
	    {
		if (rdargs)
		    rdargs->RDA_Flags |= 1 << 30;

		s1 = template;
	    }

	    if (FPuts(output, s1) || FPuts(output, ": "))
		ERROR(me->pr_Result2);

	    if (!Flush(output))
		ERROR(me->pr_Result2);

	    /* Read a line in. */
	    s1 = internal_ReadLine(cs, DOSBase);
	    if (!s1)
		ERROR(me->pr_Result2);

	    dosFreeVec(iline); iline = s1;
	}
    }

    /*
	Get enough space for string buffer.
	It's always smaller than the size of the input line + 1.
    */

    strbuf = (STRPTR)dosAllocVec((ULONG) (cs->CS_Length + 1) ,MEMF_ANY);
    if (!strbuf)
	ERROR(ERROR_NO_FREE_STORE);

    /* Count the number of items in the template (number of ',' + 1). */
    numargs = 1;
    s1 = template;

    while (*s1)
    {
	if (*s1++ == ',')
	    numargs++;
    }

    /* Use this count to get space for temporary flag array and result
       buffer. */
    flags = (UBYTE *) dosAllocVec(numargs + 1, MEMF_ANY | MEMF_CLEAR);
    argbuf = (STRPTR *) dosAllocVec((numargs + 1) * sizeof(STRPTR), MEMF_ANY | MEMF_CLEAR);

    if (!flags || !argbuf)
	ERROR(ERROR_NO_FREE_STORE);

    /* Fill the flag array. */
    s1 = template;
    s2 = flags;

    while (*s1)
    {
	/* A ',' means: goto next item. */
	if (*s1 == ',')
	    s2++;

	/* In case of a '/' use the next character as option. */
	if (*s1++ == '/')
	{
	    if (*s1 >= 'a')
		*s2 |= argflags[*s1 - 'a'];
	    else
		*s2 |= argflags[*s1 - 'A'];
	}
    }

    /* Add a dummy so that the whole line is processed. */
    *++s2 = MULTIPLE;

    /*
        Now process commandline for the first time:
        * Go from left to right and fill all items that need filling.
        * If an item is given as 'OPTION=VALUE' or 'OPTION VALUE' fill
          it out of turn.
    */
    s1 = strbuf;

    for (arg = 0; arg <= numargs; arg = nextarg)
    {
	nextarg = arg + 1;

	/* Skip /K options and options that are already done. */
	if (flags[arg] & KEYWORD || argbuf[arg] != NULL)
	    continue;

	/* Get item. Quoted items are no keywords. */
	it = ReadItem(s1, ~0ul / 2, cs);

	if (it == ITEM_UNQUOTED)
	{
	    /* Not quoted. Check if it's a keyword. */
	    item = FindArg(template, s1);
	    if (item >= 0 && item < numargs && argbuf[item] == NULL)
	    {
		/*
		** It's a keyword. Fill it and retry the current option
		** at the next turn
		*/
		nextarg = arg;
		arg = item;

		/* /S may not be given as 'OPTION=VALUE'. */
		if ((flags[item] & TYPEMASK) != SWITCH)
		{
		    /* Get value. */
		    it = ReadItem(s1, ~0ul / 2, cs);

		    /* If it was '=' then really get the value. */
		    if (it == ITEM_EQUAL)
			it = ReadItem(s1, ~0ul / 2, cs);

		    /*
		    ** If the value is missing, bail out with an error
		    ** (300702 bugfix: this test was missing - Piru)
		    ** (010203 bugfix: this test was ignoring non-/K case - Piru)
		    */
		    if (it != ITEM_UNQUOTED && it != ITEM_QUOTED)
		    {
			ERROR(ERROR_KEY_NEEDS_ARG);
		    }
		}
	    }
	}

	/* Check returncode of ReadItem(). */
	if (it == ITEM_EQUAL)
	    ERROR(ERROR_BAD_TEMPLATE);
	if (it == ITEM_ERROR)
	    ERROR(me->pr_Result2);
	if (it == ITEM_NOTHING)
	    break;

	/* /F takes all the rest. If we parsed a keyword before, 'arg' has changed
	   so we won't pass the test below. */
	if ((flags[arg] & TYPEMASK) == REST)
	{
	    argbuf[arg] = s1;

	    /* Skip the argument read by ReadItem (which may unquote it). */
	    /* The original dos unquotes the first quoted argument and leaves
	       other quotes unchanged. */
	    while (*s1++)
		;

	    --s1;
	    s2 = s1 - 1;

	    /* 310802: UnGetC last whitespace char, new ReadItem() doesn't do it
	       anymore... - Piru */
	    if (cs->CS_CurChr > 0 && (cs->CS_Buffer[cs->CS_CurChr - 1] == ' ' ||
	                              cs->CS_Buffer[cs->CS_CurChr - 1] == '\t'))
		cs->CS_CurChr--;

	    /* Find the last non-whitespace character */
	    while ((c = GETCHAR(cs)) != EOF && c != '\n')
	    {
		if (c != ' ' && c != '\t')
		 {
		    s2 = s1;
		}

		/* Copy string by the way. */
		*s1++ = (UBYTE) c;
	    }

	    /* Add terminator (1 after the character found). */
	    s2[1] = 0;
	    it = ITEM_NOTHING;
	    break;
	}

	if (flags[arg] & MULTIPLE)
	{
	    /* All /M arguments are stored in a buffer. */
	    if (multnum >= multmax)
	    {
		/* Buffer too small. Get a new one. */
		multmax += 16;
		newmult = (STRPTR *) dosAllocVec(multmax * sizeof(char *), MEMF_ANY);
		if (!newmult)
		    ERROR(ERROR_NO_FREE_STORE);

		CopyMemQuick((ULONG *) multvec, (ULONG *) newmult,
			     multnum * sizeof(char *));
		dosFreeVec(multvec);

		multvec = newmult;
	    }
	    /* Put string into the buffer. */
	    multvec[multnum++]=s1;
	    while (*s1++)
	        ;
	    /* /M takes more than one argument, so retry. */
	    nextarg=arg;
	}
	else if ((flags[arg] & TYPEMASK) == SWITCH)
	{
	    /* /S just set a flag */
	    argbuf[arg]=(char *)~0;
	}
	else if ((flags[arg] & TYPEMASK) == TOGGLE)
	{
	    /* /T set value to DOSTRUE if yes|on, or DOSFALSE if no|off */
	    /* 3rd feb 2003 bugfix: was all missing.. cool, NOT! */
	    if (Stricmp(s1, "yes") == 0 || Stricmp(s1, "on") == 0)
	    {
		argbuf[arg]=(char *) DOSTRUE;
	    }
	    else if (Stricmp(s1, "no") == 0 || Stricmp(s1, "off") == 0)
	    {
		argbuf[arg]=(char *) DOSFALSE;
	    }
	    else
	    {
		ERROR(ERROR_KEY_NEEDS_ARG);
	    }
	}
	else /* NORMAL || NUMERIC */
	{
	    /* Put argument into argument buffer. */
	    argbuf[arg]=s1;
	    while (*s1++)
	        ;
	}
    }

    /* Unfilled /A options steal Arguments from /M */
    for (arg = numargs; arg-- > 0 ; )
    {
	if (flags[arg] & REQUIRED && argbuf[arg] == NULL &&
	    !(flags[arg] & MULTIPLE))
	{
	    if (flags[arg] & KEYWORD)
	    {
		/* /K/A argument, which instists on keyword
		   being used, cannot be satisfied */
		/* yes, strange error number, but it translates
		 to "wrong number of arguments"
		 */
		ERROR(ERROR_TOO_MANY_ARGS);
	    }
	    if (!multnum)
		/* No arguments left? Oh dear! */
		ERROR(ERROR_REQUIRED_ARG_MISSING);
	    argbuf[arg]=multvec[--multnum];
	}
    }

    /* Put the rest of /M where it belongs */
    for (arg = 0; arg < numargs; arg++)
    {
	if (flags[arg] & MULTIPLE)
	{
	    if (flags[arg] & REQUIRED && !multnum)
		ERROR(ERROR_REQUIRED_ARG_MISSING);

	    if (multnum)
	    {
		/* NULL terminate it. */
		if (multnum >= multmax)
		{
		    multmax = multnum + 1;
		    newmult = (STRPTR *)dosAllocVec(multmax * sizeof(STRPTR), MEMF_ANY);
		    if (!newmult)
			ERROR(ERROR_NO_FREE_STORE);

		    CopyMemQuick((ULONG *) multvec, (ULONG *) newmult, multnum * sizeof(char *));
		    dosFreeVec(multvec);

		    multvec = newmult;
		}
		multvec[multnum++] = NULL;
		argbuf[arg] = (STRPTR)multvec;
	    }
	    else
	    {
		/* Shouldn't be necessary, but some buggy software relies on this */
		argbuf[arg] = NULL;
	    }
            break;
	}
    }

    /* There are some arguments left? Return error. */
    if (multnum && arg == numargs)
    {
	DEBUG_RDARGS(dprintf("ReadArgs: unparseable arguments left\n"));
	ERROR(ERROR_TOO_MANY_ARGS);
    }

    /*
	The commandline is processed now. Put the results in the result array.
	Convert /N arguments by the way.
    */
    for (arg = 0; arg < numargs; arg++)
    {
	/* Just for the arguments given. */
	if (argbuf[arg])
	{
	    if (flags[arg] & MULTIPLE)
	    {
		array[arg] = (IPTR)argbuf[arg];
		if ((flags[arg] & TYPEMASK) == NUMERIC)
		{
		    STRPTR *p;
		    LONG *q;

		    if (multnum * 2 > multmax)
		    {
			multmax = multnum * 2;
			newmult = (STRPTR *)dosAllocVec(multmax * sizeof(STRPTR), MEMF_ANY);
			if (!newmult)
			    ERROR(ERROR_NO_FREE_STORE);

			CopyMemQuick((ULONG *) multvec, (ULONG *) newmult, multnum * sizeof(char *));
			dosFreeVec(multvec);

			multvec = newmult;
		    }
		    array[arg] = (IPTR)multvec;
		    p = multvec;
		    q = (LONG*)(multvec+multnum);

		    while (*p)
		    {
			/* Convert /N argument. */
			chars = StrToLong(*p,q);
			if (chars <= 0 || (*p)[chars])
			{
			    /* Conversion failed. */
			    ERROR(ERROR_BAD_NUMBER);
			}
			/* Put the result where it belongs. */
			*p = (STRPTR)q;
			p++;
			q++;
		    }
                }
	    }
	    else
	    {
		switch (flags[arg] & TYPEMASK)
		{
		    case NORMAL:
		    case REST:
		    case SWITCH:
		    case TOGGLE:
			/* Simple arguments are just copied. */
			array[arg] = (IPTR) argbuf[arg];
			break;

		    case NUMERIC:
			/* Convert /N argument. */
			chars = StrToLong(argbuf[arg], &value);
			if (chars <= 0 || argbuf[arg][chars])
			{
			    /* Conversion failed. */
			    ERROR(ERROR_BAD_NUMBER);
			}
			/* Put the result where it belongs. */
#if 0
			if (flags[arg] & REQUIRED)
			{
			    /* Required argument. Return number. */
			    array[arg] = value;
			}
			else
#endif
			{
			    /* Abuse the argbuf buffer. It's not needed anymore. */
			    argbuf[arg] = (STRPTR) value;
			    array[arg] = (IPTR) &argbuf[arg];
			}
			break;
		}
	    }
	}
	else
	{
	    if (flags[arg] & MULTIPLE)
	    {
		/* Shouldn't be necessary, but some buggy software relies on this
		 * IBrowse's URL field isn't set to zero.
		 */
		array[arg] = NULL;
	    }
	}
    }

    /* All OK. */
    error = 0;
end:

    DEBUG_RDARGS(dprintf("ReadArgs: error = %ld, rda = %lx\n", error, (ULONG) rda));

    /* Cleanup and return. */
    dosFreeVec(flags);
    if (error)
    {
	/* ReadArgs() failed. Clean everything up. */
	while ((c = GETCHAR(cs)) != '\n' && c != ';' && c != EOF)
	    ;
	dosFreeVec(iline);

	if (!rdargs)
	    dosFreeVec(rda);
	dosFreeVec(dalist);
	dosFreeVec(argbuf);
	dosFreeVec(strbuf);
	dosFreeVec(multvec);
	me->pr_Result2 = error;

	return NULL;
    }
    else
    {
	/* All went well. Prepare result and return. */
	c = GETCHAR(cs);
	dosFreeVec(iline);

	rda->RDA_DAList = (IPTR) dalist;
	dalist->ArgBuf = argbuf;
	dalist->StrBuf = strbuf;
	dalist->MultVec = multvec;

	return rda;
    }
    AROS_LIBFUNC_EXIT
} /* ReadArgs */


STRPTR internal_ReadLine(struct CSource *cs,
			 struct DosLibrary *DOSBase)
{
    int isize = 0;
    int ibuf = 0;
    int c;
    STRPTR iline = NULL;
    struct Process *me = (struct Process *)FindTask(NULL);
    BPTR input = Input();

    DEBUG_RDARGS(dprintf("ReadArgs: me = $%lx, stdin = $%lx, output = $%lx\n", (ULONG) me, input, Output()));

    for (;;)
    {
        if (isize >= ibuf)
        {
	    /* Buffer too small. Get a new one. */
	    STRPTR newiline;

	    ibuf += 256;
	    newiline = (STRPTR) dosAllocVec((ULONG) (ibuf + 1), MEMF_ANY);
	    if (!newiline)
	    {
		dosFreeVec(iline);

		me->pr_Result2 = ERROR_NO_FREE_STORE;

		return NULL;
	    }
	    if (isize)
		CopyMemQuick((ULONG *) iline, (ULONG *) newiline, isize);

	    dosFreeVec(iline);

	    iline = newiline;
        }

        /* Read character */
        c = FGetC(input);
        /* Check and write it. */
        if (c == EOF && me->pr_Result2)
	{
	    dosFreeVec(iline);

	    return NULL;
	}

        if (c == EOF || c== '\n' || !c)
        {
            /* stegerg: added this. Otherwise try "list ?" then enter only "l" + RETURN
	       and you will get a broken wall in FreeMem reported. This happens in
	       FreeArgs() during the FreeVec() of the StrBuf. Appending '\n' here fixes
	       this, but maybe the real bug is somewhere else. */

            iline[isize++] = '\n';

	    /* end stegerg: */

	    break;
        }

        iline[isize++] = c;
    }

    iline[isize] = '\0';

    /* Prepare input source for new line. */
    cs->CS_Buffer = iline;
    cs->CS_Length = isize;

    return iline;
}

#ifdef TEST
#    include <dos/dos.h>
#    include <dos/rdargs.h>
#    include <utility/tagitem.h>

#define ReadArgs zzz
#    include <proto/dos.h>
#undef ReadArgs

char cmlargs[] = "T=TEST/A";

char usage[] =
   "This is exthelp for test\n"
   "Enter something";

#define CML_TEST 0
#define CML_END  1

LONG cmlvec[CML_END];

int main(int argc, char **argv)
{
    struct RDArgs *rdargs;

    if ( (rdargs = AllocDosObject(DOS_RDARGS, NULL)))
    {
	rdargs->RDA_ExtHelp = usage; /* FIX: why doesn't this work? */

	if (!(ReadArgs(cmlargs, cmlvec, rdargs)))
	{
	    PrintFault(IoErr(), "AROS boot");
	    FreeDosObject(DOS_RDARGS, rdargs);
	    exit(RETURN_FAIL);
	}
    }
    else
    {
	PrintFault(ERROR_NO_FREE_STORE, "AROS boot");
	exit(RETURN_FAIL);
    }

    FreeArgs(rdargs);
    FreeDosObject(DOS_RDARGS, rdargs);

    return 0;
} /* main */

#endif /* TEST */
